/**
 * Copyright (c) 2009-2012 EBM WebSourcing, 2012-2013 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see <http://www.gnu.org/licenses/>
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.bc.sql.su;

import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.List;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.jbi.messaging.MessagingException;
import javax.sql.DataSource;

import org.ow2.petals.bc.sql.Constants;
import org.ow2.petals.bc.sql.DataSources;
import org.ow2.petals.bc.sql.SqlComponent;
import org.ow2.petals.bc.sql.service.SQLParameter;
import org.ow2.petals.component.framework.AbstractComponent;
import org.ow2.petals.component.framework.api.configuration.ConfigurationExtensions;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.jbidescriptor.generated.Jbi;
import org.ow2.petals.component.framework.jbidescriptor.generated.Provides;
import org.ow2.petals.component.framework.su.AbstractServiceUnitManager;
import org.ow2.petals.component.framework.su.ServiceUnitDataHandler;

/**
 * Service Unit Manager for the SQL component.
 * @author Adrien Louis - EBM WebSourcing
 */
public class SUManager extends AbstractServiceUnitManager {

    private DataSources dataSources;

    /**
     * Pattern used to extract storedProcedure name from Declaration.
     */
    private static final Pattern storedProcedureNamePattern = Pattern
            .compile("^\\s*[cC][rR][eE][aA][tT][eE]\\s+[oErReEpPlLaAcC\\s]*[pP][rR][oO][cC][eE][dD][uU][rR][eE]\\s+([^\\s\\(]+)");

    /**
     * Pattern used to extract storedProcedure name from Signature.
     */
    private static final Pattern storedProcedureName2Pattern = Pattern.compile("^\\s*([^\\s\\(]+)");

    /**
     * Pattern used to extract storedProcedure parameter.
     */
    private static final Pattern paramPattern = Pattern
            .compile("^\\s*\\({0,1}\\s*([INOUT]{0,5})[@\\w]*\\s*([INOUT]{0,5})[@\\w]*\\s*([^\\(\\s,]+)[\\(\\),0-9]*\\s*([,\\)])");

    public SUManager(AbstractComponent component) {
        super(component);
    }

    @Override
    public void doDeploy(String serviceUnitName, String suRootPath, Jbi descriptor)
            throws PEtALSCDKException {

        final ServiceUnitDataHandler suDatahandler = this.getServiceUnitDataHandlers().get(
                serviceUnitName);

        if (descriptor.getServices().getConsumes().size() != 0) {
            throw new PEtALSCDKException("Consumes section are not supported by this component");
        }

        for (final Provides provides : descriptor.getServices().getProvides()) {
            final ConfigurationExtensions extensions = suDatahandler
                    .getConfigurationExtensions(provides);

            final DataSource dataSource = this.getDataSources().createDataSourceForService(
                    extensions);
            this.getDataSources().addDataSource(provides, dataSource);

            this.logger.info("DataSource for " + provides.getServiceName() + "-"
                    + provides.getEndpointName() + " created");
        }

    }

    @Override
    protected void doInit(String serviceUnitName, String suRootPath) throws PEtALSCDKException {
        final ServiceUnitDataHandler suDataHandler = this.getServiceUnitDataHandlers().get(
                serviceUnitName);
        final Jbi descriptor = suDataHandler.getDescriptor();
        final List<Provides> providesList = descriptor.getServices().getProvides();

        for (final Provides provides : providesList) {
            final ConfigurationExtensions extensions = suDataHandler
                    .getConfigurationExtensions(provides);

            try {
                if (extensions.get(Constants.STORED_PROCEDURE_CALL) == null) {
                    String storedProcedure = extensions.get(Constants.STORED_PROCEDURE);
                    if (storedProcedure != null) {
                        extractStoredProcedureElements(storedProcedure, storedProcedureNamePattern,
                                extensions, true);
                        // Register the stored statement
                        try {
                            final Connection connection = this.getDataSources().getDataSource(
                                    provides).getConnection();
                            final Statement statement = connection.createStatement();
                            statement.executeUpdate(storedProcedure);
                            this.logger.info("Stored procedure successfully registered : '"
                                    + storedProcedure + "'");
                        } catch (SQLException e) {
                            throw new PEtALSCDKException("Failed to prepare a stored procedure", e);
                        }
                    } else {
                        storedProcedure = extensions.get(Constants.STORED_PROCEDURE_SIGNATURE);
                        if (storedProcedure != null) {
                            extractStoredProcedureElements(storedProcedure,
                                    storedProcedureName2Pattern, extensions, false);
                        }
                    }
                }
            } catch (MessagingException e) {
                throw new PEtALSCDKException(e);
            }
        }
    }

    @Override
    protected void doStart(String serviceUnitName) throws PEtALSCDKException {
        final Jbi descriptor = this.getServiceUnitDataHandlers().get(serviceUnitName)
                .getDescriptor();
        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            // try to establish a connection to test if the database is
            // reachable
            try {
                final Connection connection = this.getDataSources().getDataSource(provides)
                        .getConnection();
                connection.close();
            } catch (SQLException e) {
                throw new PEtALSCDKException("Can not reach the target database", e);
            }
        }
    }

    @Override
    protected void doShutdown(String serviceUnitName) throws PEtALSCDKException {
        final ServiceUnitDataHandler suDataHandler = this.getServiceUnitDataHandlers().get(
                serviceUnitName);
        final Jbi descriptor = suDataHandler.getDescriptor();
        final List<Provides> providesList = descriptor.getServices().getProvides();
        for (final Provides provides : providesList) {
            ConfigurationExtensions extensions = suDataHandler.getConfigurationExtensions(provides);
            final String storedProcedure = extensions.get(Constants.STORED_PROCEDURE);

            if (storedProcedure != null) {
                // Unregister the stored statement
                try {
                    final Connection connection = this.getDataSources().getDataSource(provides)
                            .getConnection();
                    Statement statement = connection.createStatement();
                    statement.executeUpdate(extensions.get(Constants.STORED_PROCEDURE_DROP));
                    this.logger.info("Stored procedure successfully dropped : '"
                            + extensions.get(Constants.STORED_PROCEDURE_DROP) + "'");
                } catch (SQLException e) {
                    throw new PEtALSCDKException("Failed to drop a stored procedure", e);
                }
            }
        }
    }

    @Override
    public void doUndeploy(String serviceUnitName) throws PEtALSCDKException {
        final Jbi descriptor = this.getServiceUnitDataHandlers().get(serviceUnitName)
                .getDescriptor();
        final List<Provides> providesList = descriptor.getServices().getProvides();

        for (final Provides provides : providesList) {
            final DataSource dataSource = this.getDataSources().removeDataSource(provides);
            this.getDataSources().closeDataSource(dataSource);
            this.logger.info("DataSource for " + provides.getServiceName() + "-"
                    + provides.getEndpointName() + " closed");
        }
    }

    /**
     * No need of synchronized, as only 1 thread work with the CDK SU Manager.
     * 
     * @return
     */
    private DataSources getDataSources() {
        if (this.dataSources == null) {
            this.dataSources = ((SqlComponent) this.component).getDataSources();
        }

        return this.dataSources;
    }

    /**
     * Extract StoredProcedure elements and put them in the extensions.
     * 
     * @param storedProcedureDeclaration
     * @param storedProcedureNamepattern
     * @param extensions
     * 
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    private void extractStoredProcedureElements(String storedProcedureDeclaration,
            Pattern storedProcedureNamepattern, ConfigurationExtensions extensions,
            boolean buildDropElement) throws PEtALSCDKException, MessagingException {

        // build the stored statement call
        StringBuilder storedProcedureCall = new StringBuilder("call ");
        StringBuilder storedProcedureDrop = null;

        Matcher matcher = storedProcedureNamepattern.matcher(storedProcedureDeclaration);
        if (matcher.find()) {
            storedProcedureCall.append(matcher.group(1)).append("(");
            this.logger.finest("Found Stored Procedure name: " + matcher.group(1));
            if (buildDropElement) {
                storedProcedureDrop = new StringBuilder("DROP PROCEDURE ").append(matcher.group(1));
            }
            String trunkatedStoredProcedure = storedProcedureDeclaration.substring(matcher.end());

            short counter = 1;
            boolean isEndParameters = false;
            while (!isEndParameters) {
                matcher = paramPattern.matcher(trunkatedStoredProcedure);

                if (matcher.find()) {
                    String sqlScope = SQLParameter.DEFAULT_SCOPE;
                    if (SQLParameter.IN_SCOPE.equals(matcher.group(1))
                            || SQLParameter.OUT_SCOPE.equals(matcher.group(1))
                            || SQLParameter.INOUT_SCOPE.equals(matcher.group(1))) {
                        sqlScope = matcher.group(1);
                    } else if (SQLParameter.IN_SCOPE.equals(matcher.group(2))
                            || SQLParameter.OUT_SCOPE.equals(matcher.group(2))
                            || SQLParameter.INOUT_SCOPE.equals(matcher.group(2))) {
                        sqlScope = matcher.group(2);
                    }
                    this.logger.finest("Found a parameter with scope: " + sqlScope);
                    this.logger.finest("Found a parameter with type: " + matcher.group(3));

                    SQLParameter sqlParameter = new SQLParameter(matcher.group(3), sqlScope,
                            counter);
                    extensions.put(Constants.PARAMETER + counter, sqlParameter.toString());
                    counter++;
                    storedProcedureCall.append("?,");
                    trunkatedStoredProcedure = trunkatedStoredProcedure.substring(matcher.end());
                }
                if (matcher.group(4).equals(")")) {
                    this.logger.finest("Found end of the procedure declaration");
                    isEndParameters = true;
                }
            }
        } else {
            throw new PEtALSCDKException("Failed to identify the name of the stored procedure.");
        }

        storedProcedureCall.deleteCharAt(storedProcedureCall.length() - 1);
        storedProcedureCall.append(")");

        extensions.put(Constants.STORED_PROCEDURE_CALL, storedProcedureCall.toString());
        if (buildDropElement) {
            extensions.put(Constants.STORED_PROCEDURE_DROP, storedProcedureDrop.toString());
        }
    }
}
